(*  Title: 	Substitutions/setplus.ML
    Author: 	Martin Coen, Cambridge University Computer Laboratory
    Copyright   1993  University of Cambridge

For setplus.thy.
Properties of subsets and empty sets.
*)

open Setplus;

(*********)

(*** Rules for subsets ***)

goal Set.thy "A <= B =  (! t.t:A --> t:B)";
by (fast_tac set_cs 1);
val subset_iff = result();

goalw Setplus.thy [ssubset_def] "A < B = ((A <= B) & ~(A=B))";
by (rtac refl 1);
val ssubset_iff = result();

goal Setplus.thy "(A::'a set <= B) = ((A < B) | (A=B))";
by (SIMP_TAC (set_ss addrews [ssubset_iff]) 1);
by (fast_tac set_cs 1);
val subseteq_iff_subset_eq = result();

(*Rule in Modus Ponens style*)
goal Setplus.thy "A < B --> c:A --> c:B";
by (SIMP_TAC (set_ss addrews [ssubset_iff]) 1);
by (fast_tac set_cs 1);
val ssubsetD = result();

(*********)

goalw Setplus.thy [empty_def] "~ a : {}";
by (fast_tac set_cs 1);
val not_in_empty = result();

goalw Setplus.thy [empty_def] "(A = {}) = (ALL a.~ a:A)";
by (fast_tac (set_cs addIs [set_ext]) 1);
val empty_iff = result();


(*********)

goal Set.thy "(~A=B)  = ((? x.x:A & ~x:B) | (? x.~x:A & x:B))";
by (fast_tac (set_cs addIs [set_ext]) 1);
val not_equal_iff = result();

(*********)

val setplus_rews = [ssubset_iff,not_in_empty,empty_iff];

(*********)

val p1::p2::prems = goal HOL.thy
     "[| P ==> P-->Q;  ~P ==> ~P-->Q |] ==> Q";
by (REPEAT ((resolve_tac [p1 RS mp,p2 RS mp] 1 THEN assume_tac 1) ORELSE 
            ares_tac [excluded_middle RS disjE] 1));
val cla_case = result();

fun cla_case_tac s n = res_inst_tac [("P",s)] cla_case n;
