(*  Title: 	Substitutions/uterm.thy
    Author: 	Martin Coen, Cambridge University Computer Laboratory
    Copyright   1993  University of Cambridge

Simple term structure for unifiation.
Binary trees with leaves that are constants or variables.
*)

UTerm = Univ +

types uterm 1

arities 
  uterm          :: (term)term

consts
  UTerm		:: "(('a + nat)sexp)set => (('a + nat)sexp)set"
  Rep_UTerm	:: "'a uterm => ('a + nat)sexp"
  Abs_UTerm 	:: "('a + nat)sexp => 'a uterm"
  VAR      	:: "('a + nat)sexp => ('a + nat)sexp"
  CONST      	:: "('a + nat)sexp => ('a + nat)sexp"
  COMB     	:: "[('a + nat)sexp, ('a + nat)sexp] => ('a + nat)sexp"
  Var      	:: "'a => 'a uterm"
  Const      	:: "'a => 'a uterm"
  Comb     	:: "['a uterm, 'a uterm] => 'a uterm"
  UTerm_rec	::
    "[('a+nat)sexp, ('a+nat)sexp => 'b, ('a+nat)sexp => 'b, \
\                   [('a+nat)sexp , ('a+nat)sexp, 'b, 'b]=>'b] => 'b"
  uterm_rec	:: "['a uterm, 'a => 'b, 'a => 'b, ['a uterm, 'a uterm,'b,'b]=>'b] => 'b"

rules
  UTerm_def	"UTerm(A) == lfp(%Z. A <+> A <+> Z <*> Z)"
    (*faking a type definition...*)
  Rep_UTerm 		"Rep_UTerm(xs): UTerm(range(ATOM))"
  Rep_UTerm_inverse 	"Abs_UTerm(Rep_UTerm(xs)) = xs"
  Abs_UTerm_inverse 	"M: UTerm(range(ATOM)) ==> Rep_UTerm(Abs_UTerm(M)) = M"
     (*defining the concrete constructors*)
  VAR_def  	"VAR(v) == IN0(v)"
  CONST_def  	"CONST(v) == IN1(IN0(v))"
  COMB_def 	"COMB(t,u) == IN1(IN1(t . u))"
     (*defining the abstract constructors*)
  Var_def  	"Var(v) == Abs_UTerm(VAR(ATOM(v)))"
  Const_def  	"Const(c) == Abs_UTerm(CONST(ATOM(c)))"
  Comb_def 	"Comb(t,u) == Abs_UTerm(COMB(Rep_UTerm(t),Rep_UTerm(u)))"

     (*uterm recursion*)
  UTerm_rec_def	
   "UTerm_rec(M,b,c,d) == wfrec(trancl(pred_sexp), M, \
\                    %z g. @u. (? x. z=VAR(x) & u=b(x)) | \
\                              (? x. z=CONST(x) & u=c(x)) | \
\                              (? x y. z=COMB(x,y) & u = d(x,y,g(x),g(y))))"

  uterm_rec_def
   "uterm_rec(t,b,c,d) == \
\   UTerm_rec(Rep_UTerm(t), %x.b(Inv(ATOM,x)), %x.c(Inv(ATOM,x)), \
\                           %x y q r.d(Abs_UTerm(x),Abs_UTerm(y),q,r))"

end
